/**
 * The SceneLoader is in charge of loading scenes.
 * If need be, it may use a HTTP request to fetch the scene first.
 * If the scene is already cached, it will simply be loaded from memory.
 */
function SceneLoader() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // All currently loaded and cached scenes.
    var cachedScenes = {};

    // Error message shown when loading fails.
    var sceneLoadErrorMessage = "Scene loading failed! " +
            "This probably broke something. " +
            "Please reload the page and try again. " +
            "If that doesn't work, try clearing your browser cache (ctrl+F5 on Chrome)." +
            "If the problem persists, please do notify the creator of Kobold Adventure. " +
            "Keep in mind that it may just be your internet connection acting up!";

    // Queue for registration functions.
    var registrationQueue = [];

    // Whether or not registration processing has started.
    var registerStarted = false;

    // How many files advanced loading is currently processing.
    var loadingFiles = 0;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Sets the scroll level of the main scene, if required.
     * @param {int} scroll What to set the scroll level to.
     *                     If undefined or settings say not to scroll, will reset.
     *                     If not undefined, and settings say to scroll to save state, will scroll to the passed parameter.
     *                     If not undefined, and settings say to scroll to bottom, will scroll to bottom of the scene.
     */
    function setScroll(scroll) {
        var saveScroll = global.settings.getSaveScroll();
        var scrollBottom = global.settings.getScrollBottom();
        if (!saveScroll || typeof scroll === 'undefined') {
            global.ui.resetScroll();
        } else {
            if (scrollBottom) {
                global.ui.scrollBottom();
            } else {
                global.ui.sceneScroll = scroll;
                global.ui.setScroll();
            }
        }
    }

    /**
     * Waits ten milliseconds, then sets the scroll level.
     * Ugly fix for even uglier bugs.
     * @param {int} scroll What to set the scroll level to.
     *                     If undefined or settings say not to scroll, will reset.
     *                     If not undefined, and settings say to scroll to save state, will scroll to the passed parameter.
     *                     If not undefined, and settings say to scroll to bottom, will scroll to bottom of the scene.
     */
    function setScrollLater(scroll) {
        setTimeout(function () {
            setScroll(scroll);
        }, 10);
    }

    /**
     * Builds and returns the path to the root of the scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the root of the scene.
     */
    function buildSceneRoot(sceneName, subfolder) {
        var root = "scenes" + "/";
        if (subfolder) {
            root += subfolder + "/";
        }
        root += sceneName + "/";
        return root;
    }

    /**
     * Builds and returns a path to the main JSON file of a scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the main JSON file of the scene.
     */
    function buildJsonFilePath(sceneName, subfolder) {
        return buildSceneRoot(sceneName, subfolder) + sceneName + ".json";
    }

    /**
     * Builds and returns a path to the compiled file of a scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the compiled file of the scene.
     */
    function buildCompiledFilePath(sceneName, subfolder) {
        return "scenes/compiled/" + sceneName + "_compiled.js";
    }

    /**
     * Builds and returns a path to a JavaScript file of a scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} fileName The name of the JavaScript file.
     * @param {string} subfolder The subfolder the scene is contained in. Optional parameter.
     * @returns {string} The path to the JavaScript file.
     */
    function buildJSFilePath(sceneName, fileName, subfolder) {
        return buildSceneRoot(sceneName, subfolder) + fileName + ".js";
    }

    /**
     * Function to be called when something fails to load.
     * Shows an error-alert and hides the loading spinner.
     */
    function loadError() {
        global.error(sceneLoadErrorMessage);
    }

    /**
     * Loads the scene of which the name is passed into the game proper.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    function loadFinished(sceneName, subfolder, scrollLevel) {
        setTimeout(function () {
            that.registerStart();
            cachedScenes[sceneName].load();
            global.ui.loadbar.hide();
            setScrollLater(scrollLevel);
        }, 50);
    }

    /**
     * Loads a single JavaScript file using the advanced loading method.
     * @param {type} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {string} fileName The name of the file to load.
     */
    function loadJavaScript(sceneName, subfolder, fileName) {
        return new Promise(function (resolve, reject) {
            var script = document.createElement("script");
            script.type = `text/javascript`;
            script.onerror = reject;
            script.src = buildJSFilePath(sceneName, fileName, subfolder);
            script.onload = function () {
                loadingFiles--;
                global.ui.loadbar.update(loadingFiles);
                resolve();
            };
            document.body.appendChild(script); // Don't use jQuery for this.
        });
    }

    /**
     * Simultaneous, asynchronous JavaScript loader.
     * Also known as "advanced loading".
     * Loads all required files simultaneously, letting the browser manage download multiplicity.
     * @param {type} sceneName The name of the scene.
     * @param {array} queue The queue of files to load.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    function loadJavaScriptSimultaneously(sceneName, queue, subfolder, scrollLevel) {
        loadingFiles = queue.length;
        var promises = [];
        for (var i = 0; i < queue.length; i++) {
            promises.push(loadJavaScript(sceneName, subfolder, queue[i]));
        }

        Promise.all(promises).then(function () {
            loadFinished(sceneName, subfolder, scrollLevel);
        }, loadError);
    }

    /**
     * The developer way of loading scenes.
     * First loads the JSON, then loads all files by adding them to the DOM.
     * "Old" loading.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    function devLoad(sceneName, subfolder, scrollLevel) {
        $.getJSON(buildJsonFilePath(sceneName, subfolder))
                .done(function (data) {
                    global.ui.loadbar.show(data.length);
                    loadJavaScriptSimultaneously(sceneName, data, subfolder, scrollLevel);
                })
                .fail(loadError);
    }

    /**
     * Updates the UI when download progress happens.
     * @param {Event} evt The event triggered by the XHTML request.
     */
    function compiledLoadingProgressUpdate(evt) {
        if (evt.lengthComputable) {
            var percentComplete = Math.round((evt.loaded / evt.total) * 100);
            global.ui.loadbar.update(100 - percentComplete);
        }
    }

    /**
     * Creates an XMLHttpRequest with a registered listener that will listen to changes in download progress and update the UI accordingly.
     */
    function compiledLoadingProgressListen() {
        var xhr = new window.XMLHttpRequest();
        xhr.addEventListener("progress", compiledLoadingProgressUpdate);
        return xhr;
    }

    /**
     * Loads the compiled version of the requested scene.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    function compiledLoading(sceneName, subfolder, scrollLevel) {
        global.ui.loadbar.show(100);
        $.ajax({
            url: buildCompiledFilePath(sceneName, subfolder),
            dataType: "script",
            xhr: compiledLoadingProgressListen
        }).done(function () {
            loadFinished(sceneName, subfolder, scrollLevel);
        }).fail(loadError);
    }

    /**
     * Loads a scene via http.
     * Calls the simultaneous loading JavaScript loading function,
     * or the newer compiled loading depending on the settings.
     * Shows the loading indicator at the start.
     * @param {string} sceneName The name of the scene.
     * @param {string} subfolder The subfolder the scene can be found in. Optional parameter.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    function httpLoad(sceneName, subfolder, scrollLevel) {
        registerStarted = false;
        if (global.settings.getOldLoad()) {
            devLoad(sceneName, subfolder, scrollLevel);
        } else {
            compiledLoading(sceneName, subfolder, scrollLevel);
        }
    }

    /**
     * Processes the registration queue, emptying it one entry at a time, calling each registration function.
     * Does nothing if the registration process has not yet started.
     */
    function processRegistrationQueue() {
        if (!registerStarted) {
            return;
        }
        while (registrationQueue.length) {
            registrationQueue.shift()();
        }
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads a scene.
     * Disables all currently playing audio, if any.
     * If cached, will just call the scene's load method.
     * If not cached, will dynamically load the scene using a http call.
     * @param {string} sceneName The name of the scene to load.
     * @param {string} subfolder An optional parameter declaring the subfolder the scene is in. 
     *                           The folder "scenes" and the folder with the name of the scene itself are always implied.
     * @param {int} scrollLevel How far down to scroll the scene. Optional parameter.
     */
    this.loadScene = function (sceneName, subfolder, scrollLevel) {
        global.audio.fadeAllAudio();
        if (cachedScenes[sceneName]) {
            cachedScenes[sceneName].load();
            setScrollLater(scrollLevel);
        } else {
            httpLoad(sceneName, subfolder, scrollLevel);
        }
    };

    /**
     * Loads a tab scene.
     * Does not disable any audio.
     * Assumes the tab scene is cached.
     * @param {string} sceneName The name of the tab scene to load.
     * @param {string} subfolder An optional parameter declaring the subfolder the tab scene is in. 
     *                           The folder "scenes" and the folder with the name of the scene itself are always implied.
     */
    this.loadTabScene = function (sceneName, subfolder) {
        if (cachedScenes[sceneName]) {
            cachedScenes[sceneName].load();
        } else {
            httpLoad(sceneName, subfolder);
        }
    };

    /**
     * Registers a scene with this SceneLoader.
     * Should ideally be immediately called from the scene's JavaScript file.
     * @param {string} name The name of the loaded scene. Will be the key to which the scene is registered.
     * @param {Scene} scene An instance of the scene-object, which will be registered with the name as its key.
     */
    this.registerScene = function (name, scene) {
        cachedScenes[name] = scene;
    };


    /**
     * Queues a scene registration, and immediately processes all pending registrations if registration processessing has been started.
     * @param {function} register Registration function.
     */
    this.queueRegister = function (register) {
        registrationQueue.push(register);
        processRegistrationQueue();
    };

    /**
     * Must be called when registration of scenes can begin.
     * The scene registration queue will remain filled until this function has been called.
     */
    this.registerStart = function () {
        registerStarted = true;
        processRegistrationQueue();
    };
}
