/**
 * Lockpicking minigame.
 * @param {string} id Scene id. Used to create controllers.
 * @param {string} key Unique key-space for the lockpicking minigame.
 * @param {int} difficulty How hard the lockpicking minigame should be. Alters tumbler travel time and allowed locking times.
 * @param {int} numTumblers  How many tumblers to use. A number from one to five.
 * @param {function} onLoss Function to be called when the player loses the minigame.
 * @param {function} onWin Function to be called when the player wins the minigame.
 */
function LockPickingMinigame(id, key, difficulty, numTumblers, onLoss, onWin) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    var uniqueIdentifier = arguments.callee.name;
    var elementId = id + "_" + key + "_" + uniqueIdentifier;

    var ctrl = new ChoiceController(id, elementId);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Called when all components have finished preloading.
     */
    function preloadDone() {
        that.loadingScreen.fade();
    }

    /**
     * Build the minigame itself.
     * @param {string} components All of the built components.
     */
    function buildSelf(components) {
        var part = `<div id=${elementId} class="lockpicking">`;
        part += components;
        part += `</div>`;
        return part;
    }

    /**
     * Register the minigame's own listeners.
     */
    function ownListeners() {

    }

    // PUBLIC MEMBERS \\        this.public = "test";
    this.loadingScreen = new LockPickingLoadingScreen(id, key);
    this.helpScreen = new LockPickingHelpScreen(id, key);
    this.winScreen = new LockPickingWinScreen(id, key);
    this.loseScreen = new LockPickingLoseScreen(id, key);
    this.background = new LockPickingBackground(id, key);
    this.lockpick = new LockPickingLockPick(id, key, this);
    this.lockpickCount = new LockPickingLockPickCount(id, key);
    this.springsAndTumblers = new LockPickingSpringsAndTumblers(id, key, this, numTumblers, difficulty);

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Builds and returns the HTML.
     */
    this.build = function () {
        var components = that.loadingScreen.build();
        components += that.helpScreen.build();
        components += that.winScreen.build();
        components += that.loseScreen.build();
        components += that.background.build();
        components += that.lockpick.build();
        components += that.lockpickCount.build();
        components += that.springsAndTumblers.build();
        return buildSelf(components);
    };

    /**
     * Registers all the required listeners.
     */
    this.registerListeners = function () {
        that.loadingScreen.registerListeners();
        that.helpScreen.registerListeners();
        that.winScreen.registerListeners();
        that.loseScreen.registerListeners();
        that.background.registerListeners();
        that.lockpick.registerListeners();
        that.lockpickCount.registerListeners();
        that.springsAndTumblers.registerListeners();
        ownListeners();
    };

    /**
     * Preloads all assets required, if they have not yet already been loaded.
     */
    this.preload = function () {
        Promise.all([
            that.loadingScreen.preload(),
            that.helpScreen.preload(),
            that.winScreen.preload(),
            that.loseScreen.preload(),
            that.background.preload(),
            that.lockpick.preload(),
            that.lockpickCount.preload(),
            that.springsAndTumblers.preload()
        ]).then(preloadDone);
    };

    /**
     * Called when a tumbler has successfully finished.
     * @param {int} position Which tumbler.
     */
    this.done = function (position) {
        if (that.springsAndTumblers.done(position)) {
            that.winScreen.show();
            onWin && onWin();
        }
    };

    /**
     * Called when the user fails to set a tumbler in place.
     * @param {int} position Which tumbler.
     */
    this.fail = function () {
        global.kobold.inventory.bag.removeMisc(new Lockpick());
        that.springsAndTumblers.reset();
        setTimeout(function () {
            that.lockpick.break();
            that.lockpickCount.update();
            setTimeout(function () {
                if (!global.kobold.inventory.bag.hasMisc(new Lockpick())) {
                    that.loseScreen.show();
                    onLoss && onLoss();
                }
            }, 100);
        }, 300);
    };
}
